@@ -12,9 +12,11 @@
 // ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 // OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 
+use crate::cert::lenient_certificate_serial_number;
 use crate::der::Tag;
+use crate::signed_data::{self, SignedData};
 use crate::x509::{remember_extension, set_extension_once, Extension};
-use crate::{der, signed_data, Error, SignatureAlgorithm, Time};
+use crate::{der, Error, SignatureAlgorithm, Time};
 
 #[cfg(feature = "alloc")]
 use std::collections::HashMap;
@@ -45,11 +47,11 @@ pub trait CertRevocationList: Sealed {
 
 /// Owned representation of a RFC 5280[^1] profile Certificate Revocation List (CRL).
 ///
-/// This type is only available when using the `alloc` feature.
-///
 /// [^1]: <https://www.rfc-editor.org/rfc/rfc5280#section-5>
 #[cfg(feature = "alloc")]
+#[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
 #[allow(dead_code)] // we parse some fields we don't expose now, but may choose to expose in the future.
+#[derive(Debug, Clone)]
 pub struct OwnedCertRevocationList {
     /// A map of the revoked certificates contained in then CRL, keyed by the DER encoding
     /// of the revoked cert's serial number.
@@ -95,9 +97,10 @@ impl CertRevocationList for OwnedCertRevocationList {
 /// Borrowed representation of a RFC 5280[^1] profile Certificate Revocation List (CRL).
 ///
 /// [^1]: <https://www.rfc-editor.org/rfc/rfc5280#section-5>
+#[derive(Debug)]
 pub struct BorrowedCertRevocationList<'a> {
     /// A `SignedData` structure that can be passed to `verify_signed_data`.
-    signed_data: signed_data::SignedData<'a>,
+    signed_data: SignedData<'a>,
 
     /// Identifies the entity that has signed and issued this
     /// CRL.
@@ -126,7 +129,7 @@ impl<'a> BorrowedCertRevocationList<'a> {
                 crl_der,
                 Tag::Sequence,
                 Error::BadDer,
-                |signed_der| signed_data::parse_signed_data(signed_der, der::MAX_DER_SIZE),
+                |signed_der| SignedData::from_der(signed_der, der::MAX_DER_SIZE),
                 der::MAX_DER_SIZE,
             )
         })?;
@@ -229,9 +232,8 @@ impl<'a> BorrowedCertRevocationList<'a> {
 
     /// Convert the CRL to an [`OwnedCertRevocationList`]. This may error if any of the revoked
     /// certificates in the CRL are malformed or contain unsupported features.
-    ///
-    /// This function is only available when the "alloc" feature is enabled.
     #[cfg(feature = "alloc")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
     pub fn to_owned(&self) -> Result<OwnedCertRevocationList, Error> {
         // Parse and collect the CRL's revoked cert entries, ensuring there are no errors. With
         // the full set in-hand, create a lookup map by serial number for fast revocation checking.
@@ -347,6 +349,7 @@ impl<'a> IntoIterator for &'a BorrowedCertRevocationList<'a> {
     }
 }
 
+#[derive(Debug)]
 pub struct RevokedCerts<'a> {
     reader: untrusted::Reader<'a>,
 }
@@ -366,6 +369,7 @@ impl<'a> Iterator for RevokedCerts<'a> {
 ///
 /// [^1]: <https://www.rfc-editor.org/rfc/rfc5280#section-5>
 #[cfg(feature = "alloc")]
+#[derive(Clone, Debug)]
 pub struct OwnedRevokedCert {
     /// Serial number of the revoked certificate.
     pub serial_number: Vec<u8>,
@@ -402,6 +406,7 @@ impl OwnedRevokedCert {
 /// certificate entry.
 ///
 /// [^1]: <https://www.rfc-editor.org/rfc/rfc5280#section-5>
+#[derive(Debug)]
 pub struct BorrowedRevokedCert<'a> {
     /// Serial number of the revoked certificate.
     pub serial_number: &'a [u8],
@@ -423,9 +428,8 @@ pub struct BorrowedRevokedCert<'a> {
 
 impl<'a> BorrowedRevokedCert<'a> {
     /// Construct an owned representation of the revoked certificate.
-    ///
-    /// Only available when the "alloc" feature is enabled.
     #[cfg(feature = "alloc")]
+    #[cfg_attr(docsrs, doc(cfg(feature = "alloc")))]
     pub fn to_owned(&self) -> OwnedRevokedCert {
         OwnedRevokedCert {
             serial_number: self.serial_number.to_vec(),
@@ -446,9 +450,9 @@ impl<'a> BorrowedRevokedCert<'a> {
             //    gracefully handle such certificates.
             // Like the handling in cert.rs we choose to be lenient here, not enforcing the length
             // of a CRL revoked certificate's serial number is less than 20 octets in encoded form.
-            let serial_number = ring::io::der::positive_integer(der)
+            let serial_number = lenient_certificate_serial_number(der)
                 .map_err(|_| Error::InvalidSerialNumber)?
-                .big_endian_without_leading_zero();
+                .as_slice_less_safe();
 
             let revocation_date = der::time_choice(der)?;
 
@@ -613,10 +617,39 @@ mod tests {
             let actual = <u8 as TryInto<RevocationReason>>::try_into(*id)
                 .expect("unexpected reason code conversion error");
             assert_eq!(actual, *expected);
+            #[cfg(feature = "alloc")]
+            {
+                // revocation reasons should be Debug.
+                println!("{:?}", actual);
+            }
         }
 
         // Unsupported/unknown revocation reason codes should produce an error.
         let res = <u8 as TryInto<RevocationReason>>::try_into(7);
         assert!(matches!(res, Err(Error::UnsupportedRevocationReason)));
     }
+
+    #[test]
+    #[cfg(feature = "alloc")]
+    fn test_derived_traits() {
+        let crl = crate::crl::BorrowedCertRevocationList::from_der(include_bytes!(
+            "../tests/crls/crl.valid.der"
+        ))
+        .unwrap();
+        println!("{:?}", crl); // BorrowedCertRevocationList should be debug.
+
+        let owned_crl = crl.to_owned().unwrap();
+        println!("{:?}", owned_crl); // OwnedCertRevocationList should be debug.
+        let _ = owned_crl.clone(); // OwnedCertRevocationList should be clone.
+
+        let mut revoked_certs = crl.into_iter();
+        println!("{:?}", revoked_certs); // RevokedCert should be debug.
+
+        let revoked_cert = revoked_certs.next().unwrap().unwrap();
+        println!("{:?}", revoked_cert); // BorrowedRevokedCert should be debug.
+
+        let owned_revoked_cert = revoked_cert.to_owned();
+        println!("{:?}", owned_revoked_cert); // OwnedRevokedCert should be debug.
+        let _ = owned_revoked_cert.clone(); // OwnedRevokedCert should be clone.
+    }
 }
